@@ -1,3 +1,6 @@
+#ifndef DBXS_SEQUENCES_H
+#define DBXS_SEQUENCES_H
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -24,3 +27,5 @@ extern WTYPE next_prime(WTYPE x);
 extern int   expand_primearray_index(PrimeArray* p, int index);
 extern int   expand_primearray_value(PrimeArray* p, WTYPE value);
 extern int find_best_pair(WTYPE* basis, int basislen, WTYPE val, int adder, int* a, int* b);
+
+#endif
